local doDebug = false
--[===[@debug@
doDebug = true
--@end-debug@]===]

local folderName,addTab = ...
local emptytable = addTab.emptytable
local bossData = addTab.bossData
local L = addTab.getLocalized
local vars


-- GLOBALS: RunTime_Data
local _G = _G
local print, tonumber, math, unpack, setmetatable, table, tinsert, tContains, pairs, ipairs, next, tostring, setfenv =
	  print, tonumber, math, unpack, setmetatable, table, tinsert, tContains, pairs, ipairs, next, tostring, setfenv
local GetTime, GetInstanceInfo, GetCurrentMapAreaID, GetLFGMode, GetPartyLFGID, GetLFGCategoryForID, SetMapToCurrentZone, SendChatMessage, IsInGroup, IsPartyLFG =
      GetTime, GetInstanceInfo, GetCurrentMapAreaID, GetLFGMode, GetPartyLFGID, GetLFGCategoryForID, SetMapToCurrentZone, SendChatMessage, IsInGroup, IsPartyLFG
local DIFFICULTY_DUNGEON_NORMAL, DIFFICULTY_DUNGEON_HEROIC, LE_LFG_CATEGORY_LFD =
      DIFFICULTY_DUNGEON_NORMAL, DIFFICULTY_DUNGEON_HEROIC, LE_LFG_CATEGORY_LFD


local addon = CreateFrame("Frame","RunTime",UIParent)

local nospamtab = {}
local function nospam(t,text)
	local now = GetTime()
	if nospamtab[text] then
		if (t < 0) or (now-t < nospamtab[text]) then return end
	end
	nospamtab[text] = now
	print(("|cffff8200[RunTime Notify]|r: %s"):format(text))
end
addTab.nospam = nospam
local function debug(t,text)
	if doDebug then
		nospam(t,text)
	end
end

local function guidtype(guid)
	return tonumber(guid:sub(-14,-14))
end

local function mobid(guid)
	local gT = guidtype(guid)
	if (gT ~= 3) and (gT ~= 5) then return -1 end
	return tonumber(guid:sub(-13,-9),16)
end

local function nrGetCurrentMapAreaID()
	addon:UnregisterEvent("WORLD_MAP_UPDATE")
	SetMapToCurrentZone()
	local mapID = GetCurrentMapAreaID()
	addon:RegisterEvent("WORLD_MAP_UPDATE")
	return mapID
end

local function prettytime(t)
	return ("%02d:%06.3f"):format(math.floor(t/60),(t%60))
end

local function prettyoffset(o)
	return ("%+05.3f"):format(o)
end

local function saveloc()
	if not vars then end
	local p,r,rp,x,y = addon.display:GetPoint()
	if r and r:GetName() then
		vars.point = {p,r:GetName(),rp,x,y}
	else
		vars.point = {p,nil,rp,x,y}
	end
end

local function loadloc()
	addon:ClearAllPoints()
	if vars then
		if vars.point then
			if vars.point[2] then
				addon.display:SetPoint(vars.point[1],_G[vars.point[2]],vars.point[3],vars.point[4],vars.point[5])
			else
				addon.display:SetPoint(unpack(vars.point))
			end
		else
			vars.point = {"CENTER","UIParent","CENTER",0,0}
			loadloc()
		end
	end
end

local function showoffset(off)
	if off then
		local aOff = math.abs(off)
		if aOff >= 100 then
			addon.display.extraText:SetText(("%s%d:%05.2f"):format((off>0) and "+" or "-", math.floor(aOff/60), (aOff%60)))
		elseif aOff >= 10 then
			addon.display.extraText:SetText(("%+05.2f"):format(off))
		else
			addon.display.extraText:SetText(("%+05.3f"):format(off))
		end
		
		if off > 0 then
			addon.display.extraText:SetTextColor(1,0,0)
		elseif off < 0 then
			addon.display.extraText:SetTextColor(0,1,0)
		else
			addon.display.extraText:SetTextColor(1,0.82,0)
		end
	else
		addon.display.extraText:SetText("NO PREV")
		addon.display.extraText:SetTextColor(1,0.82,0)
	end
end

local function lfgmode()
	local slot = GetPartyLFGID()
	if not slot then return end
	local category = GetLFGCategoryForID(slot)
	if category ~= LE_LFG_CATEGORY_LFD then return end
	return GetLFGMode(LE_LFG_CATEGORY_LFD,slot)
end

local coreevents = {"LFG_UPDATE","INSTANCE_ENCOUNTER_ENGAGE_UNIT","UPDATE_INSTANCE_INFO","ADDON_LOADED","CHAT_MSG_MONSTER_YELL","COMBAT_LOG_EVENT_UNFILTERED","WORLD_MAP_UPDATE"}

local killtriggers = {}
local advkilltriggers = {}
local yelltriggers = {}
local spelltriggers = {}
local namemapping = {}
local othertriggers = {}

local function clearevents()
	table.wipe(killtriggers)
	table.wipe(yelltriggers)
	table.wipe(spelltriggers)
	table.wipe(namemapping)
	while true do
		local n,t = next(advkilltriggers,nil)
		if not n then break end
		t:Release()
		advkilltriggers[n] = nil
	end
	while true do
		local n,t = next(othertriggers,nil)
		if not n then break end
		if not tContains(coreevents,n) then addon:UnregisterEvent(n) end
		t:Release()
		othertriggers[n] = nil
	end
end

local function finishdungeon()
	local info = addon.active
	if not (addon.status == "active" and info) then return debug(0,"Attempted to finish dungeon even though no dungeon is currently active?") end
	local totalTime = GetTime()-info.startTime
	local out
	addon.display.textM:SetText(("%02d"):format(math.floor(totalTime/60)))
	addon.display.textS:SetText(("%06.3f"):format(totalTime%60))
	showoffset(curOffset)
	addon.display.extraFrame.animFrame:Hide()
	addon.display.extraFrame:Show()
	if info.prevBest then
		local offset = totalTime-info.prevBest.totalTime
		showoffset(offset)
		if offset < 0 then
			out = ("[RunTime] New best time! Dungeon %s finished. Total time is %s. (%s)"):format(info.label,prettytime(totalTime),prettyoffset(offset))
		else
			out = ("[RunTime] Dungeon %s finished. Total time is %s. Best time is %s. (%s)"):format(info.label,prettytime(totalTime),prettytime(info.prevBest.totalTime),prettyoffset(offset))
		end
	else
		out = ("[RunTime] Dungeon %s finished. Total time is %s. No previous runs recorded."):format(info.label,prettytime(totalTime))
	end
	
	if vars.announceEnd then
		SendChatMessage(out,"INSTANCE_CHAT")
	else
		nospam(0,out)
	end
	
	if (not info.prevBest) or (totalTime < info.prevBest.totalTime) then
		local bestTable
		if not info.prevBest then
			vars.best[info.mapID] = vars.best[info.mapID] or {}
			vars.best[info.mapID][info.diff] = {splits = {}}
			bestTable = vars.best[info.mapID][info.diff]
		else
			table.wipe(info.prevBest.splits)
			bestTable = info.prevBest
		end
		bestTable.totalTime = totalTime
		for id,time in pairs(info.splits) do
			bestTable.splits[id] = time
		end
	end

	addon.display.timer:Hide()
	addon.status = "finished"
	clearevents()
	addon.active:Release()
end

local function triggersplit(id)
	if id == true then return finishdungeon() end
	local info = addon.active
	if not info then return debug(0,("Attempted to trigger split for unit %s while no dungeon is active?"):format(tostring(id))) end
	if info.splits[id] then return debug(0,("Split for %s skipped - already done during this run."):format(tostring(id))) end
	local curTime = GetTime()-info.startTime
	local curOffset = (info.prevBest and info.prevBest.splits[id]) and (curTime-info.prevBest.splits[id])
	info.splits[id] = curTime
	
	addon.display.timer.animating = true
	addon.display.textM:SetText(("%02d"):format(math.floor(curTime/60)))
	addon.display.textS:SetText(("%06.3f"):format(curTime%60))
	showoffset(curOffset)
	addon.display.extraFrame.animFrame.animTimer = 0
	addon.display.extraFrame.animFrame:Show()
	addon.display.extraFrame:Show()
	
		
	local out
	if curOffset then
		out = ("[RunTime] Split: %s down. Current time is %s (%s)."):format(namemapping[id],prettytime(curTime),prettyoffset(curOffset))
	else
		out = ("[RunTime] Split: %s down. Current time is %s (%s)."):format(namemapping[id],prettytime(curTime),(info.prevBest and "Skipped on prev best" or "No prev best"))
	end
	if vars.announceSplit then
		SendChatMessage(out,"INSTANCE_CHAT")
	else
		nospam(0,out)
	end
end

local function registertriggers(id,triggers,runEnv)
	local tT = type(triggers)
	if tT == "string" then
		yelltriggers[L(triggers)] = id
	elseif tT == "number" then
		killtriggers[triggers] = id
	elseif tT == "table" then
		for _,trigger in ipairs(triggers) do
			local tType,tInfo,tInfo2 = unpack(trigger)
			if tType == "NPCDEATH" then
				killtriggers[tInfo] = id
			elseif tType == "ADVDEATH" then
				setfenv(tInfo2,runEnv)
				advkilltriggers[tInfo] = emptytable(tInfo2,id)
			elseif tType == "YELL" then
				yelltriggers[L(tInfo)] = id
			elseif tType == "SPELLCAST" then
				spelltriggers[tInfo] = id
			else
				othertriggers[tType] = othertriggers[tType] or emptytable()
				setfenv(tInfo,runEnv)
				tinsert(othertriggers[tType],emptytable(tInfo,id))
				addon:RegisterEvent(tType)
			end
		end
	end
end

local meta = {__index=setmetatable({nospam=nospam,mobid=mobid,lfgmode=lfgmode,prettytime=prettytime,prettyoffset=prettyoffset},{__index=_G})}
addon:SetScript("OnEvent",function(self,event,...)
	if event == "ADDON_LOADED" and (...) == folderName then
		RunTime_Data = RunTime_Data or {}
		vars = RunTime_Data
		vars.best = vars.best or {}
		
		loadloc()
		
		addon.status = "pending"
		
		self:UnregisterEvent("ADDON_LOADED")
	elseif event == "CHAT_MSG_MONSTER_YELL" then
		local id = yelltriggers[(...)]
		if id then
			triggersplit(id)
		end
	elseif event == "COMBAT_LOG_EVENT_UNFILTERED" then
		local _,sub,_,_,_,_,_,dst,_,_,_,id = ...
		if sub == "UNIT_DIED" then
			local dID = mobid(dst)
			local id = killtriggers[dID]
			if id then
				triggersplit(id)
			end
			local advtrig = advkilltriggers[dID]
			if advtrig and advtrig[1](event,...) then
				triggersplit(advtrig[2])
			end
		elseif sub == "SPELL_CAST_SUCCESS" then
			local id = spelltriggers[id]
			if id then
				triggersplit(id)
			end
		end
	else
		if IsInGroup() and IsPartyLFG() then
			local name, type, diff = GetInstanceInfo()
			if not name then return end
			local mapID = nrGetCurrentMapAreaID()
			if addon.status == "pending" then
				local thisData = bossData[mapID] and bossData[mapID][diff]
				if not thisData then
					return debug(-1,("No boss information found for instance '%s', ID %d, difficulty %d. Aborting."):format(name,mapID,diff))
				end
				
				local runEnv = setmetatable({},meta)
				if thisData.splits then
					for _,data in ipairs(thisData.splits) do
						local id, label, triggers = unpack(data)
						registertriggers(id,triggers,runEnv)
						namemapping[id] = label
					end
				end
				registertriggers(true,thisData.finish,runEnv)
				debug(0,("Started: %s"):format(name))
				addon.active = emptytable()
				addon.active.startTime = GetTime()
				addon.active.label = name
				addon.active.mapID = mapID
				addon.active.diff = diff
				addon.active.env = runEnv
				addon.active.splits = emptytable()
				if vars.best[mapID] and vars.best[mapID][diff] then
					addon.active.prevBest = vars.best[mapID][diff]
				end
				addon.status = "active"
				addon.startTime = GetTime()
				if diff == DIFFICULTY_DUNGEON_HEROIC then
					addon.display.header.text:SetText(("Active: %s (H)"):format(name))
				else
					addon.display.header.text:SetText(("Active: %s"):format(name))
				end
				addon.display.textS:SetText("00.000")
				addon.display.textM:SetText("00")
				addon.display.timer.s = 0
				addon.display.timer.m = 0
				addon.display.timer.animating = false
				addon.display.extraFrame.animFrame:Hide()
				addon.display.extraFrame:Hide()
				addon.display.timer:Show()
				addon.display:Show()
			end
		else
			if addon.status == "active" then
				debug(0,("Failed: %s"):format(addon.active.label))
				clearevents()
				addon.active:Release()
				addon.active = nil
				addon.status = "pending"
				addon.display:Hide()
			elseif addon.status == "finished" then
				addon.status = "pending"
				addon.display:Hide()
			end
		end
	end
	
	local other = othertriggers[event]
	if other then
		for _,info in ipairs(other) do
			if info[1](event,...) then
				triggersplit(info[2])
			end
		end
	end		
end)
for _,k in ipairs(coreevents) do addon:RegisterEvent(k) end

addon.display = CreateFrame("Frame",nil,UIParent)
addon.display:SetFrameStrata("HIGH")
addon.display:SetSize(180,60)
addon.display:SetMovable(true)
addon.display:SetBackdrop({
	bgFile="Interface\\DialogFrame\\UI-DialogBox-Background",
	edgeFile="Interface\\DialogFrame\\UI-DialogBox-Border",
	tileSize=32,
	tile=true,
	insets={
		top=12,
		bottom=12,
		right=12,
		left=12,
	},
})

addon.display.header = CreateFrame("Button",nil,addon.display)
addon.display.header:SetHeight(25)
addon.display.header.texL = addon.display.header:CreateTexture(nil,"BACKGROUND")
addon.display.header.texL:SetTexture("Interface\\AddOns\\RunTime\\tex\\capleft.tga")
addon.display.header.texL:SetPoint("TOPLEFT")
addon.display.header.texL:SetPoint("BOTTOMLEFT")
addon.display.header.texL:SetWidth(10.5)
addon.display.header.texR = addon.display.header:CreateTexture(nil,"BACKGROUND")
addon.display.header.texR:SetTexture("Interface\\AddOns\\RunTime\\tex\\capright.tga")
addon.display.header.texR:SetPoint("TOPRIGHT")
addon.display.header.texR:SetPoint("BOTTOMRIGHT")
addon.display.header.texR:SetWidth(10.5)
addon.display.header.texM = addon.display.header:CreateTexture(nil,"BACKGROUND")
addon.display.header.texM:SetTexture("Interface\\AddOns\\RunTime\\tex\\drop.tga")
addon.display.header.texM:SetPoint("TOPLEFT",10.5,0)
addon.display.header.texM:SetPoint("BOTTOMRIGHT",-10.5,0)

addon.display.header.text = addon.display.header:CreateFontString(nil,"ARTWORK")
addon.display.header.text:SetFont("Fonts\\FRIZQT__.TTF", 8)
addon.display.header.text:SetTextColor(1,1,1)

addon.display.header:SetPoint("LEFT",addon.display.header.text,"LEFT",-15,0)
addon.display.header:SetPoint("RIGHT",addon.display.header.text,"RIGHT",15,0)

addon.display.header.text:SetPoint("TOP",addon.display,"TOP",0,-5)

addon.display.header:SetScript("OnMouseDown",function() if not vars.lock then addon.display:StartMoving() end end)
addon.display.header:SetScript("OnMouseUp",function() addon.display:StopMovingOrSizing() if not vars.lock then saveloc() end end)

addon.display.textD = addon.display:CreateFontString(nil,"ARTWORK")
addon.display.textD:SetFont("Interface\\AddOns\\RunTime\\DS-DIGI.TTF",16)
addon.display.textD:SetTextColor(0.2,0.75,0.2)
addon.display.textD:SetPoint("CENTER",-10,0)
addon.display.textD:SetText(":")

addon.display.textM = addon.display:CreateFontString(nil,"ARTWORK")
addon.display.textM:SetFont("Interface\\AddOns\\RunTime\\DS-DIGI.TTF",16)
addon.display.textM:SetTextColor(0.2,0.75,0.2)
addon.display.textM:SetPoint("RIGHT",addon.display.textD,"LEFT",-3,0)

addon.display.textS = addon.display:CreateFontString(nil,"ARTWORK")
addon.display.textS:SetFont("Interface\\AddOns\\RunTime\\DS-DIGI.TTF",16)
addon.display.textS:SetTextColor(0.2,0.75,0.2)
addon.display.textS:SetPoint("LEFT",addon.display.textD,"RIGHT",3,0)

addon.display.extraFrame = CreateFrame("Frame",nil,addon.display)
addon.display.extraFrame:Hide()
addon.display.extraFrame:SetFrameLevel(addon.display:GetFrameLevel()-1)
addon.display.extraFrame:SetPoint("TOP",addon.display,"BOTTOM",0,25)
addon.display.extraFrame:SetSize(160,50)
addon.display.extraFrame:SetScale(0.5)
addon.display.extraFrame:SetBackdrop({
	bgFile="Interface\\DialogFrame\\UI-DialogBox-Background",
	edgeFile="Interface\\DialogFrame\\UI-DialogBox-Border",
	tileSize=32,
	tile=true,
	insets={
		--top=12,
		top=0,
		bottom=12,
		right=12,
		left=12,
	},
})
addon.display.extraText = addon.display.extraFrame:CreateFontString(nil,"ARTWORK")
addon.display.extraText:SetFont("Interface\\AddOns\\RunTime\\DS-DIGI.TTF",18)
addon.display.extraText:SetTextColor(0,1,0)
addon.display.extraText:SetPoint("CENTER")

addon.display:Hide()

addon.display.timer = CreateFrame("Frame",nil,addon.display)
addon.display.timer:Hide()
addon.display.timer.textM = addon.display.textM
addon.display.timer.textS = addon.display.textS
addon.display.timer:SetScript("OnUpdate",function(self,elapsed)
	self.s=self.s+elapsed
	if self.s >= 60 then
		self.m = self.m+1
		self.s = self.s-60
		if not self.animating then
			self.textM:SetText(("%02d"):format(self.m))
		end
	end
	if not self.animating then self.textS:SetText(("%06.3f"):format(self.s)) end
end)

addon.display.extraFrame.animFrame = CreateFrame("Frame",nil,addon.display.extraFrame)
addon.display.extraFrame.animFrame:Hide()
addon.display.extraFrame.animFrame.f = addon.display.extraFrame
addon.display.extraFrame.animFrame.t = addon.display.extraText
addon.display.extraFrame.animFrame.b = addon.display.timer
local pi = math.pi
addon.display.extraFrame.animFrame:SetScript("OnUpdate",function(self,elapsed)
	self.animTimer = self.animTimer+elapsed
	if self.animTimer > 7 then
		self.e = false
		self.f:Hide()
		self:Hide()
		self.b.animating = false
		self.b.textM:SetText(("%02d"):format(self.b.m))
	elseif (not self.e) and (self.animTimer >= pi) then
		self.e = true
		self.t:SetAlpha(1)
	elseif self.animTimer < pi then
		self.t:SetAlpha(0.6+math.cos(self.animTimer*4)*0.4)
	end
end)



LibStub("LibShorterSlash-1.0"):RegisterCommand({
	help={
		type="handler",
		handler=function()
			print("=== RunTime Help ===")
			print("/rt help - This help text")
			print("/rt advhelp - Help text explaining advanced commands")
			print("/rt test - Toggle test mode on/off")
			print("/rt lock - Toggle display movement on/off")
			print("/rt announce (splits||total) - Toggle /i announcement of split/total times on/off")
		end,
	},
	advhelp={
		type="handler",
		handler=function()
			print("=== RunTime Advanced Help ===")
			print("/rt invalidate - Invalidates the currently active instance and prevents it from being recorded . Can be useful if you join a run already in progress.")
			print("/rt reset - Resets the addon into its \"no run active\" state. Can be used to force detection if you requeue with an existing LFD group (teleport out of dungeon first).")
			print("/rt debuginfo - Prints debug information about the addon's current status. This information is needed to manually add new instances to the data set, among others.")
		end,
	},
	lock={
		type="handler",
		handler=function(_,arg)
			if arg:lower() == "yes" then
				vars.lock = true
			elseif arg:lower() == "no" then
				vars.lock = false
			else
				vars.lock = not vars.lock
			end
			if vars.lock then
				nospam(0,"Window is now locked.")
			else
				nospam(0,"Window is now unlocked.")
			end
		end,
	},
	announce={
		type="group",
		args={
			splits={
				type="handler",
				handler=function(_,arg)
					if arg:lower() == "yes" then
						vars.announceSplit = true
					elseif arg:lower() == "no" then
						vars.announceSplit = false
					else
						vars.announceSplit = not vars.announceSplit
					end
					if vars.announceSplit then
						nospam(0,"Split times will now be announced to /i.")
					else
						nospam(0,"Split times will now only be printed locally.")
					end
				end,
			},
			total={
				type="handler",
				handler=function(_,arg)
					if arg:lower()=="yes" then
						vars.announceEnd = true
					elseif arg:lower() == "no" then
						vars.announceEnd = false
					else
						vars.announceEnd = not vars.announceEnd
					end
					if vars.announceEnd then
						nospam(0,"Completion time will now be announced to /i.")
					else
						nospam(0,"Completion time will now only be printed locally.")
					end
				end,
			},
		},
	},
	test={
		type="handler",
		handler=function()
			if addon.status ~= "pending" then return end
			if addon.display:IsShown() then
				addon.display:Hide()
				return
			end
			addon.display.timer.m = 0
			addon.display.timer.s = 0
			addon.display.timer.animating = false
			addon.display.textM:SetText("00")
			addon.display.textS:SetText("00.000")
			addon.display.header.text:SetText("Hint: /runtime lock")
			addon.display.timer:Show()
			addon.display.extraFrame:Show()
			addon.display.extraText:SetText("(-3.42)")
			addon.display.extraText:SetTextColor(0,1,0)
			addon.display.extraText:SetAlpha(1)
			addon.display.extraFrame.animFrame:Hide()
			addon.display:Show()
		end,
	},
	invalidate={
		type="handler",
		handler=function()
			if addon.status ~= "active" then return end
			clearevents()
			addon.display:Hide()
			addon.active:Release()
			addon.active = nil
			addon.status = "finished"
		end,
	},
	reset={
		type="handler",
		handler=function()
			if addon.status ~= "finished" then return end
			addon.display:Hide()
			addon.status = "pending"
		end,
	},
	debuginfo={
		type="handler",
		handler=function()
			local name, type, diff = GetInstanceInfo()
			local mapID = nrGetCurrentMapAreaID()
			print("== RunTime debug info ==")
			print(("Addon status: %s"):format(addon.status))
			print(("Instance name: '%s' (Difficulty: %s)"):format(tostring(name),tostring(diff)))
			print(("Map ID: %s"):format(tostring(mapID)))
			if addon.active then
				print(("Active instance: '%s' (MapID %s)"):format(tostring(addon.active.label),tostring(addon.active.mapID)))
				print(("Active difficulty: %s"):format(tostring(addon.active.diff)))
			end
			print(("LFG mode: %s"):format(tostring((lfgmode()))))
			print(("Party is LFG: %s"):format(tostring(IsPartyLFG())))
			if UnitExists("target") then
				local tGUID = UnitGUID("target")
				local gT = guidtype(tGUID)
				if (gT == 0) then
					print("Target exists and is identified as a: player")
				elseif (gT == 4) then
					print("Target exists and is identified as a: pet")
				elseif (gT == 3) then
					print("Target exists and is identified as a: NPC")
				elseif (gT == 5) then
					print("Target exists and is identified as a: vehicle")
				end
				
				local tarID = mobid(tGUID)
				if tarID > 0 then
					print(("Target mob ID: %d"):format(tarID))
				end
			end
		end,
	},
},"/rt","/runtime")